/*
 * Main file
 *
 * (c) 2005 Mark Robson 
 * mark@vectrex.org.uk
 * Made for the Ludum Dare 48 hour competition, 10/12/2005
 * 
 */
#include "model.h"
#include "map.h"
#include "resourcemanager.h"
#include "gamemanager.h"
#include "player.h"
#include "playerinfo.h"
#include "constants.h"
#include "utils.h"

#include <alleggl.h>
#include <GL/glu.h>
#include <stdexcept>
#include <cmath>
#include <cstring>

static bool fullscreen = false;

PlayerInfo playerInfo;

/*
 * Rendering context states that we will never change.
 */
static void InitGl()
{
	/* remove back faces */
	glEnable(GL_CULL_FACE);
  
	/* speedups */
	glDisable(GL_DITHER);
	glShadeModel(GL_FLAT);
	glPolygonMode(GL_FRONT , GL_FILL); // GL_FILL or GL_LINE.
}

static BITMAP *hudBmp;

static void InitHud()
{
	hudBmp = create_bitmap(256, 16);
}

static void RenderHud()
{
	clear_to_color(hudBmp,0);
	int white = makecol(255,255,255);
	int black = 0;
	textprintf_ex(hudBmp, font, 1,1, white, black, 
		"Score: %05d Lives: %d Beans: %d",
		playerInfo.score, playerInfo.lives, playerInfo.beans);
	// Copy it to a texture, then render a quad.
	GLuint hudtex = allegro_gl_make_texture_ex(
		AGL_TEXTURE_FLIP,
		hudBmp,
		GL_LUMINANCE8);
	glBindTexture(GL_TEXTURE_2D, hudtex);
	// draw a quad.
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2f(0,0);
	glTexCoord2f(1,0);
	glVertex2f(512,0);
	glTexCoord2f(1,1);
	glVertex2f(512,32);
	glTexCoord2f(0,1);
	glVertex2f(0,32);
	glEnd();
	// delete the texture.
	glDeleteTextures(1, &hudtex);	
}


/*
 * Called at beginning of frame - init projection matrix.
 */
static void InitProjection(GameManager &man)
{
	glMatrixMode(GL_MODELVIEW);
  	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
  	glLoadIdentity();
	// Set the frustum...
	// Calculate Y field of view.
	float fovy = 180 * (atan((ViewHeight / 2) / ViewDistance) / M_PI);
	fovy *= 2;
	gluPerspective(fovy, (GLfloat)ScreenWidth/(GLfloat)ScreenHeight, 
		ViewDistance / 2,
		ViewDistance * 3);
	glTranslatef(- man.viewCentre.x, - man.viewCentre.y, - ViewDistance - man.viewCentre.z);
		
	glMatrixMode(GL_MODELVIEW);
 	
	// glEnable(GL_LIGHTING);
	static GLfloat LightPosition[] = {1.0f, 1.5f, 1.0f, 0.0f}; // last number 0 = Infinity distance in that dir

	static GLfloat LightAmbient[]  = {0.4f, 0.4f, 0.4f, 1.0f};
	static GLfloat LightDiffuse[]  = {0.6f, 0.6f, 0.6f, 1.0f};
	glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);
	glLightfv(GL_LIGHT1, GL_POSITION, LightPosition);
	glEnable(GL_LIGHT1);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_NORMALIZE); // Normalise our normals please.
	
	glDepthFunc(GL_LESS);
	glClearDepth(ViewDistance * 3);
}

static void SetOrthoProjection()
{
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
  	glLoadIdentity();
	glOrtho(0, ScreenWidth, 0, ScreenHeight, -1, 1);
	glMatrixMode(GL_MODELVIEW);
}

static void InitRender(GameManager &man)
{
	// Render background, clear buf etc..
	glClearColor(0,0,0.6, 0);
	glClear(GL_DEPTH_BUFFER_BIT);
	// Render background.
	SetOrthoProjection();
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glBindTexture(GL_TEXTURE_2D, 
		resourceManager.getTexture(ResourceManager::BackgroundTexture));
	// draw a quad.
	double textureOffset = (man.viewCentre.x) / (TileSize * 50);
	glBegin(GL_QUADS);
	glTexCoord2f(0 + textureOffset,0);
	glVertex2f(0,0);
	glTexCoord2f(1 + textureOffset,0);
	glVertex2f(ScreenWidth,0);
	glTexCoord2f(1 + textureOffset,1);
	glVertex2f(ScreenWidth,ScreenHeight);
	glTexCoord2f(0 + textureOffset,1);
	glVertex2f(0,ScreenHeight);
	glEnd();
	
	
}

static void InitAllegro()
{
	allegro_init();
	install_allegro_gl();

	allegro_gl_clear_settings();
	/* allegro_gl_set (AGL_Z_DEPTH, 16); */
	allegro_gl_set (AGL_COLOR_DEPTH, 16);
	allegro_gl_set (AGL_DOUBLEBUFFER, 1);
	allegro_gl_set (AGL_Z_DEPTH, 16);
	// TODO: Make fullscreen an option.
	if (fullscreen) {
		allegro_gl_set (AGL_FULLSCREEN, TRUE);	
	} else {
		allegro_gl_set (AGL_WINDOWED, TRUE);	
	}
	allegro_gl_set (AGL_SUGGEST, AGL_COLOR_DEPTH | AGL_Z_DEPTH
	              | AGL_DOUBLEBUFFER | AGL_WINDOWED);

	set_color_depth (16);
	if (set_gfx_mode(GFX_OPENGL, ScreenWidth, ScreenHeight, 0, 0) < 0) {
		throw std::runtime_error("Unable to set gfx mode");
	}
	int depthBits;
	glGetIntegerv(GL_DEPTH_BITS, & depthBits);
	if (depthBits == 0) {
		throw std::runtime_error("No depth buffer");
	}
	install_keyboard();
	install_mouse();
	install_timer();
	install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, "");
	InitGl();
}

static const int NumModels = 4;

// Init things which only need to be done once.
static void InitAll()
{
	InitAllegro();
	InitHud();
	
	resourceManager.loadAll();
}

static void Render(GameManager &man)
{
	InitRender(man);
	InitProjection(man);
	// DO main rendeirng here.
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	man.render();
	// ModelTest();
	// Set ortho for text.
	SetOrthoProjection();
	// Render HUD etc here.
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
	RenderHud();
	allegro_gl_flip();
}

void LevelLoop(GameManager &man)
{
	while ((! playerInfo.levelComplete) && (playerInfo.deadTime < DeathTimeout)) {
		man.tick();
		if (key[KEY_LEFT]) {
			man.viewOffset.x -= 0.1;
		}
		if (key[KEY_RIGHT]) {
			man.viewOffset.x += 0.1;
		}
		if (key[KEY_UP]) {
			man.viewOffset.y += 0.1;
		}
		if (key[KEY_DOWN]) {
			man.viewOffset.y -= 0.1;
		}
		Render(man);
		rest(20);
		if (key[KEY_ESC]) {
			break;
		}
	}
}

void GameLoop()
{
	playerInfo = PlayerInfo();
	GameManager man;
	
	do {
		man.cleanUp();
		char levelFilename[200];
		std::sprintf(levelFilename, "maps/map%d.txt", playerInfo.level);
		theMap.load(levelFilename, &man);
		char msg[200];
		sprintf(msg, "Entering level:%d", playerInfo.level);
		ShowIntermission(msg);
		LevelLoop(man);
		if (playerInfo.dead) {
			playerInfo.nextLife();
		}
		if (playerInfo.levelComplete) {
			playerInfo.nextLevel();
		}
		if (key[KEY_ESC]) {
			break;
		}
	} while ((playerInfo.lives > 0) && (playerInfo.level <= NumLevels));
	if (playerInfo.level > NumLevels) {
		char msg[200];
		sprintf(msg, "You win, score:%d", playerInfo.score);
		ShowIntermission(msg);
	} else {
		ShowIntermission("Game over");
	}
}

static void MainLoop()
{
	while (ShowTitle()) {
		GameLoop();
	}
}

int main(int argc, char *argv[])
{
	try {
		if (argc>1) {
			if (argv[1][0] == 'f') {
				fullscreen = true;
			}
		}
		InitAll();
		MainLoop();
	} catch (std::runtime_error &e) 
	{
		allegro_message ("Unexpected exception:%s\nAllegro error: %s\n"
		                 "Allegro GL error : %s\n",
		                 e.what(), allegro_error, allegro_gl_error);
	} catch (...) {
		allegro_message ("Unknown exception thrown");
	}
	set_gfx_mode(GFX_TEXT, 0,0, 0, 0);
	return 0;
}

END_OF_MAIN()


