
#include "map.h"
#include "resourcemanager.h"
#include "gamemanager.h"
#include "player.h"
#include "enemies.h"

#include <GL/glu.h>
#include <GL/gl.h>
// Allegro - used to read the level bitmap.
#include <allegro.h>
#include <stdexcept>

#include <fstream>
#include <iostream>

// Global.
Map theMap;

static const double SpikeScale = 0.3;
static const double DoorScale = 0.8;

Map::Map()
{
	// Blank it.
	data.resize(MapHeight);
	for (int i=0; i<MapHeight; i++) {
		data[i].resize(MapWidth);
	}
	clear();

}

void Map::clear()
{
	for (int i=0; i<MapHeight; i++) {
		fill(data[i].begin(), data[i].end(), Tile());
	}
	// Make the floor and ceiling.
	for (int i=0; i<MapWidth; i++) {
		tileAt(i, 0).type = Wall;	
		tileAt(i, MapHeight-1).type = Wall;
	}
}

void Map::render(int minx, int miny, int maxx, int maxy)
{
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	glColor3f(1,1,1);
	for (int x=minx; x <=maxx; x++) {
		for (int y=miny; y <=maxy; y++) {
			renderBlock(x,y);
		}
	}
}

void Map::renderObject(int x, int y, ResourceManager::ModelType modelType, double scale)
{
	glBindTexture(GL_TEXTURE_2D, 
		resourceManager.getTexture(ResourceManager::GenericTexture));
	const Model &model = resourceManager.getModel(modelType);
	glPushMatrix();
	glTranslatef(x * TileSize + (TileSize / 2), y * TileSize, 0);
	glScalef(scale, scale, scale);
	model.draw(true);
	glPopMatrix();
}

/*
 * Render a backdrop piece.
 */
void Map::renderWall(int x, int y)
{
	double leftx = x * TileSize;
	double bottomy = y * TileSize;
	double rightx = leftx + TileSize;
	double topy = bottomy + TileSize;
	int textureId = resourceManager.getTexture(ResourceManager::StoneTexture);
	glBindTexture(GL_TEXTURE_2D, textureId);
	glBegin(GL_QUADS); // GL_QUADS
	// Backdrop.
	glNormal3f(0,0,1);
	glTexCoord2d(1,0);
	glVertex3d(rightx, bottomy, -(TileDepth / 2));
	glTexCoord2d(1,1);
	glVertex3d(rightx, topy, -(TileDepth / 2));
	glTexCoord2d(0,1);
	glVertex3d(leftx, topy, -(TileDepth / 2));
	glTexCoord2d(0,0);
	glVertex3d(leftx, bottomy, -(TileDepth / 2));
	glEnd();
}

void Map::renderBlock(int x, int y)
{
	// Find out what's there
	// decide whether to do anything
	Tile t = tileAt(x,y);
	ResourceManager::Texture txt;
	
	if (t.indoors) {
		renderWall(x,y);
	}
	switch (t.type) {
		case Blank:
			return;
		case Spikes:
			renderObject(x,y, ResourceManager::Spikes, SpikeScale);
			return;
		case Door1:
			renderObject(x,y, ResourceManager::Door1, DoorScale);
			return;
		case Door2:
			renderObject(x,y, ResourceManager::Door2, DoorScale);
			return;
		case Wall:
			txt = ResourceManager::BrickTexture;
			break;
		case Earth:
			txt = ResourceManager::EarthTexture;
			break;
		default:
			return;
	}
	// Draw the pieces of the wall.
	
	double leftx = x * TileSize;
	double bottomy = y * TileSize;
	double rightx = leftx + TileSize;
	double topy = bottomy + TileSize;
	int textureId = resourceManager.getTexture(txt);
	
	glBindTexture(GL_TEXTURE_2D, textureId);
	glBegin(GL_QUADS); // GL_QUADS
	// Front of the wall.
	glNormal3f(0,0,1);
	glTexCoord2d(1,0);
	glVertex3d(rightx, bottomy, (TileDepth / 2));
	glTexCoord2d(1,1);
	glVertex3d(rightx, topy, (TileDepth / 2));
	glTexCoord2d(0,1);
	glVertex3d(leftx, topy, (TileDepth / 2));
	glTexCoord2d(0,0);
	glVertex3d(leftx, bottomy, (TileDepth / 2));
	// Left of the wall. optimisation: check if there is anything to the left, if so, omit.
	
	glNormal3f(-1,0,0);
	glTexCoord2d(0,0);
	glVertex3d(leftx, bottomy, (TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(leftx, topy, TileDepth / 2);
	glTexCoord2d(1,1);
	glVertex3d(leftx, topy, -(TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(leftx, bottomy, -(TileDepth / 2));
	// Right
	glNormal3f(1,0,0);
	glTexCoord2d(0,0);
	glVertex3d(rightx, bottomy, (TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(rightx, bottomy, -(TileDepth / 2));
	glTexCoord2d(1,1);
	glVertex3d(rightx, topy, -(TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(rightx, topy, TileDepth / 2);
	// Top
	glNormal3f(0,1,0);
	glTexCoord2d(0,0);
	glVertex3d(leftx, topy, (TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(rightx, topy, (TileDepth / 2));
	glTexCoord2d(1,1);
	glVertex3d(rightx, topy, - (TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(leftx, topy, - (TileDepth / 2));
	
	
	// Bottom.
	glNormal3f(0,-1,0);
	glTexCoord2d(0,0);
	glVertex3d(leftx, bottomy, (TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(leftx, bottomy, - (TileDepth / 2));
	glTexCoord2d(1,1);
	glVertex3d(rightx, bottomy, - (TileDepth / 2));
	glTexCoord2d(1,0);
	glVertex3d(rightx, bottomy, (TileDepth / 2));	
	glEnd();
}

static void createItemAt(GameManager *man, int x, int y, GameObject *obj)
{
	man->add(obj);
	obj->location.x = (x * TileSize) + (TileSize / 2);
	obj->location.y = (y * TileSize);
}

void Map::load(const std::string &filename, GameManager *man)
{
	clear();
	std::ifstream in(filename.c_str());
	if (! in.good()) {
		throw std::runtime_error("Can't load map:" + filename);
	}
	// Ignore first two lines (future use)
	in.ignore(1000, '\n');
	in.ignore(1000, '\n');
	int y = MapHeight - 1;
	bool finished = false;
	while ((! finished) && (y > 0)) {
		char buf[150];
		in.getline(buf,150);
		int len = std::strlen(buf);
		// std::cout << "[" << buf << "]" << endl;
		if (len ==0) {
			finished =true;
		}
		for (int x =0; x < len; x++) {
			Tile &t = tileAt(x,y);
			t.indoors = true;
			switch (buf[x]) {
				case '.':
					t.type = Blank;
					t.indoors = true;
					break;
				case '#':
					t.type = Wall;
					break;
				case 'E':
					t.type = Earth;
					break;
				case 'd':
					t.type = Door1;
					break;
				case 'D':
					t.type = Door2;
					break;
				case '!':
					t.type = Spikes;
					break;
				case 't':
					createItemAt(man, x,y, new Treasure());
					break;
				case 'b':
					createItemAt(man, x,y, new Bean());
					break;
				case 'j':
					{
						GameObject *p = new Player();
						createItemAt(man, x,y, p);
						man->setTrack(p);
						break;
					}
				case ' ':
					t.indoors = false;
					break;
				case 'g':
					createItemAt(man, x,y, new Giant());
					break;
			}
		}
		y --; // move down a row.
	}
}


/*
// CRAP OLD LOADER.
// red bits:
// 0-63 : empty
// 64-127 : indoor
// 128-191 : soil
// 192+ : brick.
void Map::load(const std::string &filename)
{
	clear();
	// Don't convert anything thanks very much.
	set_color_conversion(COLORCONV_NONE);
	PALETTE pal;
	BITMAP *bmp = load_bitmap(filename.c_str(), pal);
	if (bmp == 0) {
		throw std::runtime_error("Can't load map:" + filename);
	}
	if ((bmp->w > MapWidth) || (bmp->h > MapHeight)) {
		throw std::runtime_error("Map too big:" + filename);
	}
	// Go from the bottom to the top.
	for (int j=0; j< bmp->h; j++) {
		int pixely = bmp->h - j;
		for (int i=0; i < bmp->w; i++) {
			int pixel = getpixel(bmp, i, pixely);
			int red = getr32(pixel);
			int green = getg32(pixel);
			int blue = getb32(pixel);
			int num = red / 64;
			
			tileAt(i, j).indoors = false;
			switch(num) {
				case 0 :
					tileAt(i, j).type = Blank;
					break;
				case 1 :
					tileAt(i, j).type = Blank;
					tileAt(i, j).indoors = true;
					if (green > 128) {
						tileAt(i, j).type = Spikes;
					} else {
						if (blue > 128) {
							tileAt(i, j).type = Door1;
						}
						if (blue > 192) {
							tileAt(i, j).type = Door2;
						}
					}
					break;
				case 2 :
					tileAt(i, j).type = Earth;
					break;
				case 3 :
					tileAt(i, j).type = Wall;
					break;
			}
		}
	}
	destroy_bitmap(bmp);
}
*/


