#include "player.h"
#include "map.h"
#include "stalk.h"

#include "playerinfo.h"

#include <GL/gl.h>
#include <allegro.h>
#include <cmath>

static const float PlayerScale = 0.2;
static const float WalkAccel = 0.01;
static const float MaxWalkSpeed = 0.10;
static const float FloorFriction = 0.005;
static const float AnimMultiplier = 1.5;

static const float ClimbSpeed = 0.05;
static const float GrabSpeed = 0.10; // Speed at which we grab the stalk

static const float GravityAccel = 0.01;
static const float FallMaxSpeed = 1.0;
static const float DeathSpeed = 0.6;
static const float JumpSpeed = 0.26;
static const float ThrowSpeed = 0.05;

Player::Player()
{
	azimuth = 0;
	desiredAzimuth = 0;
	anim = 0;
	frame = ResourceManager::JackWalking1;
	nextFrame = frame;
	state = Standing;
	stateTimer = 1000;
	radius = 1.0;
	height = 2.5;
	fireTimer = 0;
}

void Player::render()
{
	const Model &from = resourceManager.getModel(frame);
	const Model &to = resourceManager.getModel(nextFrame);
	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	if (state == Dead)  {
		glBindTexture(GL_TEXTURE_2D, 
			resourceManager.getTexture(ResourceManager::PlayerDeadSkin));
	} else {
		glBindTexture(GL_TEXTURE_2D, 
			resourceManager.getTexture(ResourceManager::PlayerSkin));
	}
	glColor3f(1.0, 1.0,1.0);
	glPushMatrix();
	glTranslatef(location.x, location.y, location.z);
	glRotatef(azimuth, 0,1,0);
	glScalef( PlayerScale, PlayerScale, PlayerScale);
	
	from.drawTweened(true, to, anim);	
	glPopMatrix();
}

void ClippedObject::processFriction()
{
	if (onGround) {
		if (std::fabs(velocity.x) < FloorFriction) {
			velocity.x = 0;
		} else {
			if (velocity.x > 0) {
				velocity.x -= FloorFriction;
			} else {
				velocity.x += FloorFriction;
			}
		}
	}
}

void Player::doFire()
{
	
	if (playerInfo.beans == 0) return;
	playerInfo.beans --;
	Bean *b = new Bean();
	b->location = location;
	b->location.y += height / 2; // start half way up Jack's body.
	b->velocity = velocity;
	// Chuck it out.
	b->velocity.x += std::sin(toRadians(azimuth)) * ThrowSpeed;
	b->velocity.y += (ThrowSpeed / 2);
	manager->add(b);
	manager->triggerSound(ResourceManager::Throw, location);
}

void Player::processInput()
{
	bool onStalk = theMap.isClimbable(location.x, location.y + 0.1);
	if (onGround) {
		if (key[KEY_A]) {
			desiredAzimuth = -90;
			state = Walking;
			stateTimer = 50;
			velocity.x -= WalkAccel;
		}
		if (key[KEY_D]) {
			desiredAzimuth = 90;
			state = Walking;
			stateTimer = 50;
			velocity.x += WalkAccel;
		}
		if (velocity.x > MaxWalkSpeed) {
			velocity.x = MaxWalkSpeed;
		}
		if (velocity.x < -MaxWalkSpeed) {
			velocity.x = -MaxWalkSpeed;
		}
		if (key[KEY_W] && !onStalk) {
			velocity.y = JumpSpeed;
		}
	}
	if (fireTimer > 0) {
		fireTimer --;
	} else {
		if (key[KEY_SPACE]) {
			// Fire a bean.
			doFire();
			fireTimer = 40;
		}
	}
	if (onStalk) {
		float climb = 0;
		if (key[KEY_W]) {
			climb = ClimbSpeed;
			state = Climbing;
		}
		if (key[KEY_S]) {
			climb = - ClimbSpeed;
			state = Climbing;
		}
		if (! onGround) {
			if (key[KEY_A]) {
				// Stop climbing.
				state = Walking;
				velocity.x = - MaxWalkSpeed;
			}
			if (key[KEY_D]) {
				// Stop climbing.
				state = Walking;
				velocity.x = MaxWalkSpeed;
			}
		}
		if (state == Climbing) {
			float offset = std::fmod(location.x, (double) TileSize);
			if (offset > (TileSize * 0.55)) {
				velocity.x = - GrabSpeed;
			} else {
				if (offset < (TileSize * 0.45)) {
					velocity.x = GrabSpeed;
				} else {
					velocity.x = 0;
				}
			}
			velocity.y = climb;
		}
	} else {
		if (state == Climbing) {
			state = Walking;
		}
	}
	// Blatant cheat.
	if (key[KEY_F12]) {
		playerInfo.levelComplete = true;
	}
}

void ClippedObject::clip()
{
	// Check if we're below the floor.
	// If so, move us up.
	if (theMap.isSolid(location.x, location.y)) {
		location.y = TileSize * std::ceil(location.y / TileSize);
		if (velocity.y < 0) {
			velocity.y = 0;
		}
	}
	float rightx = location.x + radius;
	float leftx = location.x - radius;
	// Left
	if (theMap.isSolid(leftx, location.y) || (theMap.isSolid(leftx, location.y + height))) {
		if (velocity.x < 0) {
			velocity.x = 0;
			// Move us out of the object.
			leftx = TileSize * std::ceil(leftx / TileSize);
			location.x = leftx + radius;
		}
	}
	// Right
	if (theMap.isSolid(rightx, location.y) || (theMap.isSolid(rightx, location.y + height))) {
		if (velocity.x > 0) {
			velocity.x = 0;
			// Move us out of the object.
			rightx = TileSize * std::floor(rightx / TileSize);
			location.x = rightx - radius;
		}
	}
	// Top
	float topy = location.y + height;
	if (theMap.isSolid(location.x, topy)) {
		if (velocity.y > 0) {
			velocity.y = 0;
			// Move us out of the object.
			topy = TileSize * std::floor(topy / TileSize);
			location.y = topy - height;
		}
	}
	onGround = theMap.isSolid(location.x, location.y - 0.1);
}

void Player::checkDeath()
{
	// Check whether we've hit some spikes or what.
	if (theMap.isSpiky(location.x, location.y + (height / 2))) {
		state = Dead;
	}
	// Check whether we've won.
	// yay won!
	Tile t = theMap.tileAtLocation(location.x, location.y);
	if (t.type == Door2) {
		playerInfo.levelComplete = true;
	}
}

void Player::tick()
{
	// Sort out velocity
	processFriction();
	if (state != Dead) {
		processInput();
	} else {
		playerInfo.dead = true;
		playerInfo.deadTime ++;
	}	
	// Apply gravity. Gravity doesn't affect us while climbing.
	if (state != Climbing) {
		velocity = velocity + Vect3d(0, - GravityAccel, 0);
		if (velocity.y < (-FallMaxSpeed)) {
			velocity.y = -FallMaxSpeed;
		}
	}
	
	location = location + velocity;
	clip();
	checkDeath();
	
	double animSpeed;
	switch (state) {
		case Walking:
			if (onGround) {
				animSpeed = std::fabs(velocity.x) * AnimMultiplier;
			} else {
				animSpeed = 0; // in mid-air
			}
			break;
		case Dead:
			if (state == Dead) {
				animSpeed = onGround ? 0.05 : 0;
			} else {
				animSpeed = 0.1;
			}
			break;
		case Climbing:
			animSpeed = velocity.y * AnimMultiplier;
			desiredAzimuth = 180;
			break;
		default:
			animSpeed = 0.1;
	}
	
	// Animate. vary animation speed with speed.
	anim += animSpeed;
	if (anim < 0) {
		anim += 1.0;
		setNextFrame();
	}
	if (anim > 1.0) {
		anim -= 1.0;
		setNextFrame();	
	}
	// Only if we're on the ground.
	if (onGround && (state != Dead)) {
		if (stateTimer > 0) {
			stateTimer --;
			if (stateTimer == 0) {
				state = Standing;
				desiredAzimuth = 0;
			}
		}
	}

	// Turn around to the desired direction
	if (std::fabs(azimuth - desiredAzimuth) < 12) {
		azimuth = desiredAzimuth;
	} else {
		if (azimuth < desiredAzimuth) {
			azimuth += 8;
		} else {
			azimuth -= 8;
		}
	}
}

void Player::setNextFrame()
{
	// Ideally this should do something useful.
	// This should handle things other than walking too.
	frame = nextFrame;
	switch (state) {
		case Walking:
			// If we're walking,
			nextFrame = (ResourceManager::ModelType) (frame + 1);
			if (nextFrame > ResourceManager::JackWalking4) {
				nextFrame = ResourceManager::JackWalking1;
			}
			if (nextFrame == ResourceManager::JackWalking2) {
				manager->triggerSound(ResourceManager::Step1, location);
			}
			if (nextFrame == ResourceManager::JackWalking4) {
				manager->triggerSound(ResourceManager::Step2, location);
			}
			break;
		case Climbing:
			if (frame == ResourceManager::JackClimbing1) {
				nextFrame = ResourceManager::JackClimbing2;
			} else {
				nextFrame = ResourceManager::JackClimbing1;
			}
			break;
		case Standing:
			nextFrame = ResourceManager::Jack;
			break;
		case Dead:
			if (nextFrame != ResourceManager::JackDead)
			{
				manager->triggerSound(ResourceManager::Death, location);
			}
			nextFrame = ResourceManager::JackDead;
			break;
	}
}

// Check for collision with enemies.
void Player::collided(GameObject * other)
{
	// Kill player.
	if (other->deadly) {
		state = Dead;
	}
}


Bean::Bean()
{
	radius = 0.25;
	height = 0.5;
	buryDepth = 0;
}

void Bean::render()
{
	// Billboard.
	glBindTexture(GL_TEXTURE_2D, 
		resourceManager.getTexture(ResourceManager::BeanTexture));
	// Set blending
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Disable lighting.
	glDisable(GL_LIGHTING);
	glTexCoord2d(1,0);
	float leftx = location.x - radius;
	float rightx = location.x + radius;
	float bottomy = location.y - buryDepth;
	float topy = location.y  - buryDepth + height;
	glBegin(GL_QUADS);
	glVertex3d(rightx, bottomy, location.z);
	glTexCoord2d(1,1);
	glVertex3d(rightx, topy, location.z);
	glTexCoord2d(0,1);
	glVertex3d(leftx, topy, location.z);
	glTexCoord2d(0,0);
	glVertex3d(leftx, bottomy, location.z);
	glEnd();
	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);		
}

void Bean::tick()
{
	velocity = velocity + Vect3d(0, - GravityAccel, 0);
	if (velocity.y < (-FallMaxSpeed)) {
		velocity.y = -FallMaxSpeed;
	}
	location = location + velocity;
	clip();
	// Apply some friction.
	processFriction();
	if (onGround && velocity.x == 0) {
		// Check if we're on earth...
		if (theMap.isEarth(location.x, location.y - 0.1)) {
			buryDepth += 0.02;
			if (buryDepth >= height) {
				growStalk();
			}
		} else {
			// Enable this to be collidable, so that we can pick it up.
			collidable = true;
		}
	}
}

void Bean::growStalk()
{
	// Use up the earth. If it's already used, don't grow another plant.
	Tile & t = theMap.tileAtLocation(location.x, location.y -0.1);
	if (t.earthUsed) {
		return;
	} else {
		t.earthUsed = true;
	}
	// Grow into a stalk.
	dead = true;
	Stalk *stalk = new Stalk();
	stalk->location = location;
	// Move to the centre of the tile.
	stalk->location.x = TileSize * (std::floor(location.x / TileSize)) + (TileSize /2 );
	// Move back z-axis slightly so Jack can climb it.
	stalk->location.z -= (TileSize / 2);
	manager->add(stalk);
	// Make an entertainign noise.
	manager->triggerSound(ResourceManager::Growing, location);
}

// Can only collide with the player.
void Bean::collided(GameObject * other)
{
	dead = true;
	playerInfo.beans ++;
}



Treasure::Treasure()
{
	radius = 0.5;
	height = 1.0;
	collidable = true;
}

void Treasure::render()
{
	// Billboard.
	glBindTexture(GL_TEXTURE_2D, 
		resourceManager.getTexture(ResourceManager::TreasureTexture));
	// Set blending
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Disable lighting.
	glDisable(GL_LIGHTING);
	glTexCoord2d(1,0);
	float leftx = location.x - radius;
	float rightx = location.x + radius;
	float bottomy = location.y;
	float topy = location.y + height;
	glBegin(GL_QUADS);
	glVertex3d(rightx, bottomy, location.z);
	glTexCoord2d(1,1);
	glVertex3d(rightx, topy, location.z);
	glTexCoord2d(0,1);
	glVertex3d(leftx, topy, location.z);
	glTexCoord2d(0,0);
	glVertex3d(leftx, bottomy, location.z);
	glEnd();
	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);		
}


// Can only collide with the player.
void Treasure::collided(GameObject * other)
{
	dead = true;
	playerInfo.score += 100;
}





